class MissingDocIdException(Exception):
    """Raise when document id is missing"""

    def __init__(self, message: str):
        super().__init__(message)


class EmptyDocListException(Exception):
    """Raise when document list is empty"""

    def __init__(self, message: str):
        super().__init__(message)



from typing import List
from collections import defaultdict
from rankum.models import Doc
from rankum.readers import DocReader


class RankDiversifier():
    """ Diversifying rank strategies"""

    def __init__(self, reader: DocReader):
        self.reader = reader

    def diversify(self, by: str) -> List[Doc]:
        """
        Diversifying document list based on a target
        :param by (str): criteria (target) to diversify by. It is an attribute common to all documents
        :return diversified rank (List[Doc]): original document list re-ranked by diversified target
       """
        raise NotImplemented


class ScoresDiffDiversifier(RankDiversifier):
    """
    This algorithm calculates the final document score based on its global relevance (original rank position) and
    local relevance (rank position in the group we are diversifying by).
    This algorithm is proposed in this paper:
    https://www.researchgate.net/publication/266658487_Using_score_differences_for_search_result_diversification
    """

    def __init__(self, reader: DocReader):
        super().__init__(reader)
        self.iter_docs = reader.to_doc_list()

    def diversify(self, by: str) -> List[Doc]:
        """
        Diversifying document list using Scores diff strategy
        :param by (str): criteria (target) to diversify by. It is an attribute common to all documents
        :return diversified rank (List[Doc]): original document list re-ranked by diversified target
        """
        local_rank = defaultdict(int)
        global_rank = []

        for i, _doc in enumerate(self.iter_docs):
            target = _doc.get(by)
            local_score = 0 if target is None else 1.0 / (local_rank.get(target, 0) + 1)
            global_score = 1.0 / (i + 1)
            _doc.score = round(local_score + global_score, 3)

            local_rank[target] += 1
            global_rank.append(_doc)

        return sorted(global_rank, reverse=True)



from functools import total_ordering


@total_ordering
class Doc:
    """Represents a document to be re-ranked"""

    def __init__(self, id: str, **kwargs):
        self.id = id
        self.__dict__.update(kwargs)

    def set_score(self, score: float):
        """
        Set a document score
        :param score (float): document score
        :return None
        """
        self.score = score

    def get(self, key: str):
        """
        Get a document field
        :param key (str): document field name
        :return document field value
        """
        return self.__dict__.get(key)

    def __eq__(self, other):
        return self.id == other.id and self.__dict__ == other.__dict__

    def __lt__(self, other):
        return self.__dict__.get('score', 0) < other.__dict__.get('score', 0)

    def __repr__(self):
        return f'Doc(id={self.id}, {self.__dict__})'



from typing import List, Iterator
from rankum.models import Doc
from rankum.exceptions import MissingDocIdException, EmptyDocListException

import json


class DocReader:
    """ Reader document list from different sources"""

    def to_doc_list(self) -> Iterator[Doc]:
        """
        Convert document list to Iterator[Doc]
        :return doc iterator (Iterator[Doc]): iterator for converted document list
        """
        raise NotImplemented()


class JsonDocReader(DocReader):
    """ Reader document list from a JSON string """

    def __init__(self, json_str: str):
        self.json_str = json_str

    def to_doc_list(self) -> Iterator[Doc]:
        """
        Convert document list to Iterator[Doc]
        :return doc iterator (Iterator[Doc]): iterator for converted document list
        """
        doc_dict = self._json_to_dict()

        if len(doc_dict) == 0:
            raise EmptyDocListException(message='Document list is empty')

        def iterator_over(all_docs: List):
            for _doc in all_docs:
                if 'id' not in _doc:
                    raise MissingDocIdException(message='Field "id" is mandatory for each document')

                yield Doc(id=str(_doc.pop('id')), **_doc)

        return iterator_over(doc_dict)

    def _json_to_dict(self) -> dict:
        """
        Convert json string to Dictionary
        :return:
        """
        try:
            doc_dict = json.loads(self.json_str)
        except TypeError as e:
            raise EmptyDocListException(message='Document list is invalid')

        return doc_dict
